require 'json'
require 'fileutils'
require 'digest/md5'
require 'pp'
require 'uri'
require 'net/http'

# Class for managing metadata about CycleCloud pools, mostly to consolidate
# common code patterns for access. If you add a method here, add it to the
# usage section in README.rdoc

# rubocop:disable Style/VariableName
# rubocop:disable Style/ClassVars

class CycleCloudCluster
  # Keep a cache of all the stuff we query from the blackboard so we don't
  # download files multiple times per converge
  @@node_cache = {}
  @discoverable_data_stored = false

  # Internal only
  def self.init
    return unless @ran_init.nil?
    @ran_init = true

    @ohai = Ohai::System.new
    @ohai.all_plugins

    @cached_role = if File.exist? cache_file
                     JSON.parse(File.read(cache_file))
                   else
                     {}
                   end
  end

  # Internal Only
  def self.cache_file
    init
    return "#{@node[:cyclecloud][:home]}\\config\\cyclecloud_cache.json" if @ohai[:os] == 'windows'
    "#{@node[:cyclecloud][:home]}/config/cyclecloud_cache.json"
  end

  def reset_node_cache
    @@node_cache = {}
  end

  #####################
  ##### NEW STUFF #####
  #####################

  def escape(str)
    Chef::Log.debug "Escaping '#{str}'"
    '+-!(){}[]^"~*?:'.each_char do |c|
      str = str.gsub(c, "\\#{c}")
    end
    str
  end

  def initialize(node)
    @node = node
  end

  def use_from_hash?
    # We have to explicitly cast the json as a Chef::Node object assuming it's properly formatted json
    # Use from_hash for >= 12.7, and json_create otherwise
    fromhash_version = '12.7.2'.split('.').map(&:to_i)
    current_version = Chef::VERSION.split('.').map(&:to_i)

    # UFO operator <=>, compares an array element by element.
    # Returns 0 if equal, -1 if lesser, 1 if greater.
    (current_version <=> fromhash_version) >= 0
  end

  def blackboard_config
    results = {}
    raise 'Invalid chef-solo blackboard configuration!' if @node[:cyclecloud][:blackboard][:type].nil?
    results['scheme'] = @node[:cyclecloud][:blackboard][:type]
    raise 'Invalid scheme in blackboard location data' if results['scheme'] != 's3' && results['scheme'] != 'cs'
    if results['scheme'] == 's3'
      Chef::Log.debug 'Detected AWS S3 scheme for blackboard location'
      # IMPORTANT: S3 URIs in us-east may not be valid FQDNs so we use the low
      # level URI.split here
      uri = URI.split(@node[:cyclecloud][:blackboard][:url]).compact
      results['bucket'] = uri[1]
      results['access_key'] = @node[:cyclecloud][:blackboard][:access_key]
      results['secret_key'] = @node[:cyclecloud][:blackboard][:secret_key]
      results['detect_credentials'] = @node[:cyclecloud][:blackboard][:detect_credentials]
      results['default_region'] = @node[:cyclecloud][:blackboard][:default_region]
      prefix = uri[2]
      prefix[0] = '' if prefix
      prefix += '/' if prefix && (!prefix.end_with? '/')
      results['prefix'] = prefix
    elsif results['scheme'] == 'cs'
      uri = URI(@node[:cyclecloud][:blackboard][:url])
      results['host'] = uri.host
      results['port'] = uri.port
      results['path'] = uri.path
      results['use_https'] = uri.scheme.casecmp('https') == 0
      results['username'] = @node[:cyclecloud][:blackboard][:username]
      results['password'] = @node[:cyclecloud][:blackboard][:password]
    end
    results
  end

  def _create_pogo_ini(config)
    Dir.mkdir("#{@node[:cyclecloud][:home]}/config") unless File.exist?("#{@node[:cyclecloud][:home]}/config")
    config_path = File.join("#{@node[:cyclecloud][:home]}/config", 'blackboard-pogo.ini')

    config_content = <<-EOS
[pogo blackboard]
type = s3
matches = s3://#{config['bucket']}/
access_key = #{config['access_key']}
secret_key = #{config['secret_key']}
use_instance_creds = #{config['detect_credentials']}
default_region = #{config['default_region']}
    EOS

    File.write(config_path, config_content)
    config_path
  end

  def serialize_node
    # TODO: Do we need any of the changes it makes?
    # The from_hash method here ensures we create a deep copy
    # otherwise the .delete method below modifies the global node object
    node_hash = Chef::Node.from_hash(@node.for_json).for_json

    to_trim_from_json = %w(filesystem filesystem2 packages dmi cpu block_device)
    to_trim_from_json.each do |key|
      unless node_hash['automatic'].delete(key).nil?
        Chef::Log.debug "Trimmed key #{key} from node data"
      end
    end

    node_hash = ::CycleCloud::Helpers.force_encoding_utf8(node_hash)

    # construct Chef::Node to ensure we haven't trimmed necessary fields
    # this is a fatal error that will require a code change, but I want to ensure it is obvious what
    # needs to change when this issue does arise. -rh
    begin
      if use_from_hash?
        Chef::Node.from_hash(node_hash)
      else
        Chef::Node.json_create(node_hash)
      end
    rescue => e
      Chef::Log.error e.to_s
      raise 'Trimming node json data is failing Chef::Node.from_hash - reconsider removing the ' \
            "following keys from node[:automatic] - #{to_trim_from_json}"
    end

    JSON.dump(node_hash)
  end

  def store_discoverable(node_data = serialize_node)
    config = blackboard_config

    #node_data = serialize_node

    ### Search using AWS S3 ###
    if config['scheme'] == 's3'

      store_discoverable_pogo(node_data)

    ### Search using CycleServer ###
    elsif config['scheme'] == 'cs'

      store_discoverable_cs(node_data)

      Chef::Log.info 'Uploaded searchable data to CycleServer...'
    else
      raise 'Invalid blackboard configuration for searchable recipe'
    end

    if @discoverable_data_stored == true
      # Storing discoverable node attrs multiple times generally indicates that
      # node states are changing w.r.t. search, so we need to disable the cache
      # (simply resetting/invalidating is not sufficient since nodes are asynch)
      Chef::Log.info 'Auto-Disabling local search caching.'
      @node.override['cyclecloud']['search']['caching']['disabled'] = true
      reset_node_cache
    end
    @discoverable_data_stored = true
  end

  def store_discoverable_cs(node_data)
    Chef::Log.debug 'Storing discoverable node record to server...'
    config = blackboard_config

    clusterUID = @node[:cyclecloud][:cluster][:id]
    instanceID = @node[:cyclecloud][:instance][:id]
    accountUID = 'default'

    http = Net::HTTP.new(config['host'], config['port'])
    if config['use_https']
      http.use_ssl = true
      if @node[:cyclecloud][:skip_ssl_validation]
        http.verify_mode = OpenSSL::SSL::VERIFY_NONE
      end
    end

    retries = 5
    begin
      url = URI.escape("/cloud/chefnode/#{accountUID}/#{clusterUID}/#{instanceID}")
      
       # If a context path is set, prepend it to the URL
      if config['path'] and config['path'] != "" and config['path'] != "/"
        url = config['path'] + url
      end
      initheader = { 'Content-Type' => 'application/json' }
      req = Net::HTTP::Put.new(url, initheader)
      req.basic_auth(config['username'], config['password'])
      req.body = node_data
      response = http.start { |http_conn| http_conn.request(req) }
    rescue
      raise unless (retries -= 1) > 0
      Chef::Log.warn 'Unable to post search data to CycleCloud - retrying in 30 seconds.'
      sleep 30
      retry
    end
    raise "Unable to post search data to CycleCloud: (#{response.code}) #{response.message}" if response.code.to_i >= 300

    # TODO: Check response status and fail if it's bad....
    # response.code, response.message, response.body
  end

  def store_discoverable_pogo(node_data)
    Chef::Log.debug 'Storing discoverable node record using pogo...'
    config = blackboard_config
    clusterUID = @node[:cyclecloud][:cluster][:id]
    instanceID = @node[:cyclecloud][:instance][:id]

    key = "#{config['prefix']}#{clusterUID}/#{instanceID}.json"

    blackboard_url = "s3://#{config['bucket']}/#{key}"

    config_path = _create_pogo_ini(config)

    # Create a temporary cache that will be removed after the search is done
    Dir.mktmpdir do |dir|
      node_data_path = File.join(dir, "#{instanceID}.json")
      File.write(node_data_path, node_data)
      cmd = "#{pogo_bin} --config=#{config_path} put #{node_data_path} #{blackboard_url} --force"
      Chef::Log.info "Running : #{cmd}"
      system cmd
    end

    Chef::Log.info "Uploaded searchable data to blackboard as #{key}"
  end

  def fetch_discoverable(options, nodes)
    cluster_uid = options[:clusterUID]

    if !(@node['cyclecloud']['search']['caching']['disabled']) && @@node_cache.key?(cluster_uid)
      Chef::Log.debug "Using cache blackboard results for clusterUID='#{cluster_uid}'"
      nodes.concat @@node_cache[cluster_uid]
      return nodes
    end

    discovered_nodes = []

    config = blackboard_config

    if config['scheme'] == 's3'

      ### Do a S3 blackboard search using prefix search and pulling out .json files ###
      fetch_discoverable_pogo(cluster_uid, discovered_nodes)

    elsif config['scheme'] == 'cs'

      fetch_discoverable_cs(cluster_uid, discovered_nodes)

    else
      raise 'Unknown blackboard scheme, unable to perform search!'
    end

    unless @node['cyclecloud']['search']['caching']['disabled']
      @@node_cache[cluster_uid] = discovered_nodes
    end

    nodes.concat discovered_nodes
    nodes
  end

  def pogo_bin
    if @node[:os] == 'windows'
      'C:/cycle/jetpack/bin/pogo.cmd'
    else
      'pogo'
    end
  end

  def fetch_discoverable_pogo(clusterUID, nodes)
    Chef::Log.debug 'Fetching discoverable node records using pogo...'
    config = blackboard_config

    key_prefix = config['prefix']
    key_prefix += clusterUID if clusterUID != :all

    blackboard_url = "s3://#{config['bucket']}/#{key_prefix}"

    config_path = _create_pogo_ini(config)

    # Create a temporary cache that will be removed after the search is done
    Dir.mktmpdir do |dir|
      cmd = "#{pogo_bin} --config=#{config_path} sync #{blackboard_url} #{dir} --delete-removed"
      Chef::Log.info "Running : #{cmd}"
      system cmd

      # Load each result (that ends in .json) into the nodes array for searching
      Dir.glob(File.join(dir, '*.json')).each do |filename|
        Chef::Log.info "Processing discoverable node: #{filename}"
        node_data = JSON.parse(File.read(filename))
        node_data = Mash.new(node_data) if node_data.is_a? Hash

        node_data = if use_from_hash? && node_data['json_class'] == 'Chef::Node'
                      Chef::Node.from_hash(node_data)
                    else
                      Chef::Node.json_create(node_data)
                    end

        nodes << node_data
      end
    end

    Chef::Log.info "Returning #{nodes.length} nodes"
    nodes
  end

  def fetch_discoverable_cs(clusterUID, nodes)
    Chef::Log.debug 'Fetching discoverable node records from server...'
    config = blackboard_config

    customerUID = 'default'
    uri = "/cloud/chefnode/#{customerUID}/"
    uri += "#{clusterUID}/" if clusterUID != :all

    # If a context path is defined, prepend it to the URI
    if config['path'] and config['path'] != "" and config['path'] != "/"
      uri = config['path'] + uri
    end

    http = Net::HTTP.new(config['host'], config['port'])
    if config['use_https']
      http.use_ssl = true
      if @node[:cyclecloud][:skip_ssl_validation]
        http.verify_mode = OpenSSL::SSL::VERIFY_NONE
      end
    end

    req = Net::HTTP::Get.new(URI.escape(uri))
    req.basic_auth(config['username'], config['password'])
    response = http.start { |http_conn| http_conn.request(req) }
    data = JSON.parse(response.body)
    # NOTE: We get a list back from CycleServer
    data.each do |node_data|
      node_data = Mash.new(node_data) if node_data.is_a? Hash
      node_data = if use_from_hash? && node_data['json_class'] == 'Chef::Node'
                    Chef::Node.from_hash(node_data)
                  else
                    Chef::Node.json_create(node_data)
                  end

      nodes << node_data
    end

    nodes
  end

  def role_missing?(node, options)
    role_no_match = options.include?(:role) && !node[:roles].include?(options[:role])
    if role_no_match 
      Chef::Log.debug "Node role reject. #{options[:role]} not found in #{node[:roles]}"
    end
    role_no_match
  end

  def recipe_missing?(node, options)
    recipe_no_match = options.include?(:recipe) && !node[:recipes].include?(options[:recipe])
    if recipe_no_match
      Chef::Log.debug "Node recipe reject. #{options[:recipe]} not found in #{node[:recipes]}"
    end
    recipe_no_match
  end

  def search(opts={}, &block)
    options = {
      clusterUID: @node[:cyclecloud][:cluster][:id], # clusterID should default to the cluster the current node is in
      singular: nil,   # If set to a string, will be thrown if exactly one result is not found
      retries: 5,      # Number of retries if an exception occurs
      retry_delay: 10  # Initial delay between retries (doubles every failed attempt)
    }.merge(opts)

    retries = 0

    # Make sure we have a clusterUID
    if options[:clusterUID].nil?
      options[:clusterUID] = @node[:cyclecloud][:cluster][:id]
    end

    begin
      customerUID = 'default'

      query_text = "ec2_userdata_user_data_customer_uid:#{escape(customerUID)}"
      if options[:clusterUID] != :all
        clusterUID = options[:clusterUID]
        query_text << " AND ec2_userdata_user_data_pool_uid:#{escape(clusterUID)}"
      end

      if options.include? :role
        # filter by role
        query_text << " AND roles:#{escape(options[:role])}"
      end

      if options.include? :recipe
        # filter by recipe
        query_text << " AND recipes:#{escape(options[:recipe])}"
      end

      nodes = []
      if !(Chef::Config[:solo])
        webserver = @node[:cyclecloud][:config][:web_server]
        webserver.slice!('http://')
        webserver.slice!('https://')
        webserver = escape(webserver)
        webserver = "*#{webserver}"

        query_str = "ec2_userdata_user_data_config_webServer:#{webserver}"
        query_str << " AND ec2_userdata_user_data_customer_uid:#{escape(customerUID)}"
        if options[:clusterUID] != :all
          clusterUID = options[:clusterUID]
          query_str << " AND ec2_userdata_user_data_pool_uid:#{escape(clusterUID)}"
        end

        if options.include? :role
          # filter by role
          query_str << " AND roles:#{escape(options[:role])}"
        end

        if options.include? :recipe
          # filter by recipe
          # query_str << " AND recipes:#{options[:recipe].gsub(':','\\:')}"
          query_str << " AND recipes:#{escape(options[:recipe])}"
        end

        Chef::Log.debug "Running chef search query: #{query_str}"
        query = Chef::Search::Query.new
        results = []
        query.search(:node, query_str) do |anode|
          results << anode
        end
        Chef::Log.debug "Results count: #{results.length}"

      else # We are using chef solo, so do a blackboard search

        nodes = fetch_discoverable(options, nodes)

        Chef::Log.info "Found #{nodes.length} nodes"

        results = nodes.reject{|node| role_missing?(node, options) || recipe_missing?(node, options)}

        if block
          results = results.select {|node|  block.call(node)}
        end

      end
      
      if results.empty?
        Chef::Log.debug("No results found for search!")
      end

      # return everything
      return results unless options[:singular]
      # return the single result
      return results[0] unless results.count != 1
      # Throw an error if we didn't get exactly 1 result
      raise SearchException, "#{options[:singular]}: search expected 1 result but found #{results.count} results for #{query_text}"
    rescue => e
      if e.instance_of? SearchException
        # simple error
        Chef::Log.warn e.to_s
      else
        Chef::Log.warn e.inspect
      end

      raise unless retries < options[:retries]
      Chef::Log.info "Search failed, retrying in #{options[:retry_delay]} seconds..."
      # Invalidate the cache to force a full fetch (otherwise retries are pointless)
      reset_node_cache
      sleep(options[:retry_delay])
      options[:retry_delay] *= 2
      retries += 1
      retry
    end
  end

  def scheduler_ip
    @scheduler_ip ||= get_role('scheduler', 'ipv4')
  end

  def cm_hostname
    @cm_hostname ||= get_role('central_manager', 'hostname')
  end

  def cm_ip
    @cm_ip ||= get_role('central_manager', 'ipv4')
  end

  def scheduler
    @scheduler ||= get_role('scheduler')
  end

  def parent?
    @parent ||= !(search(clusterUID: :all) do |anode|
                    next if anode[:cyclecloud][:cluster].nil?
                    anode[:cyclecloud][:cluster][:parent_id] == @node[:cyclecloud][:cluster][:id]
                  end
                 ).empty?
  end

  def child?
    @child ||= @node[:cyclecloud][:cluster][:parent_id].nil?
  end

  # Given a recipe name and optionally a attribute and clusterUID, find the appropriate node in the
  # cluster and return the attribute based on the provider
  def get_recipe(recipe, attribute = 'hostname', clusterUID = nil, singular = nil)
    Chef::Log.debug "get_recipe: #{recipe} #{attribute} #{clusterUID}"

    attribute_value = nil

    recipes = @node[:recipes]
    if recipes.include?(recipe) && clusterUID.nil?
      Chef::Log.debug "ME HOSTNAME: #{recipe} #{attribute} #{clusterUID}"
      attribute_value = @node['cyclecloud']['instance'][attribute]
    else
      result = if !clusterUID.nil?
                 search(recipe: recipe, clusterUID: clusterUID, singular: singular)
               else
                 search(recipe: recipe, singular: singular)
               end
      result = result.first if result.is_a? Array

      unless result.nil?
        if result['cyclecloud'] && result['cyclecloud']['instance']
          attribute_value = result['cyclecloud']['instance'][attribute]
        end
      end
    end

    attribute_value
  end

  def get_role(role, attribute = 'hostname', clusterUID = nil, singular = nil)
    Chef::Log.debug "get_role: #{role} #{attribute} #{clusterUID}"

    attribute_value = nil

    if @node[:roles].include?(role) && clusterUID.nil?
      Chef::Log.debug "ME HOSTNAME: #{role} #{attribute} #{attribute_value}"
      attribute_value = @node['cyclecloud']['instance'][attribute]
    else # no it must be on a different host in this cluster
      result = if !clusterUID.nil?
                 search(role: role, clusterUID: clusterUID, singular: singular)
               else
                 search(role: role, singular: singular)
               end

      result = result.first if result.is_a? Array

      if result['cyclecloud'] && result['cyclecloud']['instance'] && !result.nil?
        attribute_value = result['cyclecloud']['instance'][attribute]
        attribute_value = result['cyclecloud']['instance'][attribute.to_sym] if attribute_value.nil?
      end

    end

    attribute_value
  end
  # this line is a little tricky. The guard keeps the class from being reloaded after it has
  # been mocked in rspec tests
end unless defined?(CycleCloudCluster)

# Thrown for "not found"-type errors rather than unexpected errors (connection problems)
class SearchException < RuntimeError
end

# Add a cluster method to all recipes which allows us to easily use a CycleCloudCluster instance properly instantiated
class Chef
  class Recipe
    @cluster = nil
    def cluster
      @cluster = CycleCloudCluster.new(node) if @cluster.nil?
      @cluster
    end
  end
end 
