#!/opt/cycle/jetpack/system/embedded/bin/python
import argparse
import datetime
import json
import logging
import logging.config
import os
import random
import sys
import traceback
import platform

import jetpack
import jetpack.util.cloudinit
import jetpack.admin

import jetpack.nodehealthchecks
from jetpack.status import JetpackStatus
from jetpack.util import UserError, jetpack_home
from jetpack.util.system import SystemUtilityError
from jetpack.util.system import copy_file, create_group, is_systemd, service, modify_file, execute, windows_env, create_user

INSTALLATION_COMPLETE_PATH = os.path.join(jetpack.util.jetpack_home(), 'run', 'jetpack.installed')
MAX_CONFIGURATION_ATTEMPTS = 5
DEFAULT_FORCE_NODEHEALTHCHECKS_ERROR_EVERY = 2

# Jetpack import can break installer logging (particularly on windows), so configure explicitly
cyclecloud_home = jetpack.util.jetpack_home()
logfile = os.path.join(cyclecloud_home, 'logs', 'install.log')
logger = None


def setup_logging(loglevel="INFO", logfile=logfile):

    LOGGING = {
        'version': 1,
        'disable_existing_loggers': False,
        'formatters': {
            'default': {
                'format': '%(asctime)s %(levelname)-8s %(message)s'
            },
        },
        'handlers': {
            'file': {
                # The values below are popped from this dictionary and
                # used to create the handler, set the handler's level and
                # its formatter.
                '()': 'logging.handlers.RotatingFileHandler',
                'level': loglevel.upper(),
                'formatter': 'default',
                # The values below are passed to the handler creator callable
                # as keyword arguments.
                'filename': logfile,
            },
        },
        'root': {
            'handlers': ['file'],
            'level': loglevel.upper(),
        },
    }

    logging.config.dictConfig(LOGGING)
    return logging.getLogger(__name__) 


def _filepath(f):
    return os.path.abspath(os.path.join(os.path.dirname(__file__), 'files', f))


def install(args):
    logger.info("Installing Jetpack %s..." % jetpack.util.version())

    # for testing remotely
    if args.fail_every != -1:
        if random.randint(0, args.fail_every - 1) == 0:
            raise UserError("Deliberate installation error occurred", detail="This is randomly chosen if the node has ForceJetpackErrors set to true.", 
                            recommendation="Set ForceJetpackErrors to false on the node and try again")

    if jetpack.util.is_linux():

        create_group('cyclecloud')

        # NOTE: We do not "need" a cyclecloud user here (unused in Jetpack), this is only to get around an issue with
        # a prebaked image (w/ jetpack) and waagent. Because the `cyclecloud` group is created above,
        # waagent will not be able to create the `cyclecloud` user add add them to an already existing group.
        # By pre-creating an unused `cyclecloud` user, waagent will skip the user creation step and avoid the
        # issue.
        create_user('cyclecloud', group='cyclecloud')
        copy_file(_filepath('cyclecloud'), '/etc/sudoers.d/cyclecloud', 'root', 'root', 0o400)
        copy_file(_filepath('cyclecloud.sh'), '/etc/profile.d/cyclecloud.sh', 'root', 'root', 0o644)

        if is_systemd():
            copy_file(_filepath('jetpackd.service'), '/etc/systemd/system/jetpackd.service', 'root', 'root', 0o644)
            copy_file(_filepath('healthcheck.service'), '/etc/systemd/system/healthcheck.service', 'root', 'root', 0o644)
        else:
            copy_file(_filepath('jetpackd.sysvinit'), '/etc/init.d/jetpackd', 'root', 'root', 0o755)
            copy_file(_filepath('healthcheck.sysvinit'), '/etc/init.d/healthcheck', 'root', 'root', 0o755)

        # Ensure all the executable scripts have correct permissions
        for f in ['bin/jetpack', 'system/bin/cyclecloud-env.sh', 'system/bin/jetpack_wrapper.sh', 'system/bootstrap/cron_wrapper.sh']:
            p = os.path.join(jetpack.util.jetpack_home(), f)
            modify_file(p, permissions=0o750)

        # Ensure the jetpack directory has proper ownership
        execute('chown -R root:cyclecloud /opt/cycle/jetpack', shell=True)

        # Ensure proper directory permissions
        for f in ['', 'bin', 'config', 'logs', 'system']:
            p = os.path.join(jetpack.util.jetpack_home(), f)
            modify_file(p, permissions=0o750)

    else:
        windows_env('CYCLECLOUD_HOME', jetpack.util.jetpack_home())
        windows_env('CYCLECLOUD_BOOTSTRAP', os.path.join(jetpack.util.jetpack_home(), 'system', 'bootstrap'))
        windows_env('PATH', '%PATH%;{0}'.format(os.path.join(jetpack.util.jetpack_home(), 'bin')))
        execute(r'sc create jetpackd binPath= "C:\cycle\jetpack\system\embedded\python\python.exe C:\cycle\jetpack\system\embedded\python\scripts\jetpackd"', shell=True, msg="Created 'jetpackd' service")
        execute('sc description jetpackd "Manages a CycleCloud enabled VM"', shell=True, msg="Set description for 'jetpackd' service")
        service('jetpackd', enable=True)


if __name__ == "__main__":
    logger = setup_logging()
    
    parser = argparse.ArgumentParser(description="Install CycleCloud Jetpack")
    parser.add_argument('--connection', help="Encoded connection string to validate route to CycleCloud")
    parser.add_argument('--connection-file', help="JSON formatted connection file to validate route to CycleCloud")
    parser.add_argument('--cloud-init', dest='cloudinit', action='store_true', help="Wait for cloud init to finish before starting Jetpack services")

    # This is a hidden flag used to force failures
    parser.add_argument('--fail-every', dest="fail_every", type=int, default=-1, help=argparse.SUPPRESS)

    # Deprecated flags, here just to prevent errors
    parser.add_argument('--batch', action='store_true', help=argparse.SUPPRESS)  # Not used anymore - just here to prevent errors if specified by an older CC
    parser.add_argument('--start-services', dest='start_services', action='store_true', help=argparse.SUPPRESS)
    
    args = parser.parse_args()

    logger.info("Args: {}".format(str(args)))

    jetpack.util.set_cyclecloud_env()

    result = {}
    exception_result = {}
    try:
        try:
            install(args)
        except SystemUtilityError as e:
            raise UserError(str(e))

        # Initialize jetpack (enable/start services)
        # TODO: Move cloud-init and connection test to initialize
        if args.connection or args.connection_file:
            jetpack.admin.initialize(connection=args.connection, connection_file=args.connection_file, cloudinit=args.cloudinit)
          
        # make sure this is called with appropriate "template" (aka nodearr name) value
        configs = jetpack.util.get_current_config()
        nodehealthchecks_forced_failures_enabled = False
        try:
            nodehealthchecks_enabled = configs['cyclecloud']['node']['nodearray_parameters']['enable_nodehealthchecks']
            cluster_session_id = configs['cyclecloud']['node']['nodearray_parameters']['cluster_session_id']
            if 'force_nodehealthchecks_errors' in configs['cyclecloud']['node']['nodearray_parameters']:
                nodehealthchecks_forced_failures_enabled = configs['cyclecloud']['node']['nodearray_parameters']["force_nodehealthchecks_errors"]
                if 'force_nodehealthchecks_errors_every' in configs['cyclecloud']['node']['nodearray_parameters']:
                    nodehealthchecks_fail_every = configs['cyclecloud']['node']['nodearray_parameters']["force_nodehealthchecks_errors_every"]
                else:
                    nodehealthchecks_fail_every = DEFAULT_FORCE_NODEHEALTHCHECKS_ERROR_EVERY
        except KeyError:
            nodehealthchecks_enabled = False

        # we want to run healthchecks before jetpackd is started for this to be useful in overprovisioning
        if nodehealthchecks_enabled:
            if nodehealthchecks_forced_failures_enabled:
                if random.randint(0, nodehealthchecks_fail_every - 1) == 0:
                    jetpack.nodehealthchecks.healthcheck(fail_test=True, cluster_session_id=cluster_session_id)
            else:
                jetpack.nodehealthchecks.healthcheck(cluster_session_id=cluster_session_id)

        jetpack.admin.start_service()

        # Write the successful installation file to disk
        with open(INSTALLATION_COMPLETE_PATH, 'w') as f:
            f.write(datetime.datetime.now().isoformat())
  
        result["status"] = JetpackStatus.Ok
        logger.info("Installation complete")
        # Writing to stderr may be treated as a failure by WAAgent on Windows
        # and stdout must be a parsable JSON object returned to CC
        if not platform.system().lower().startswith("win"):
            sys.stderr.write("Installation complete. See %s for details.\n" % logfile)

    except UserError as e:
        logger.exception("Error installing Jetpack")
        exception_result["status"] = JetpackStatus.Error
        exception_result["message"] = str(e)
        exception_result["description"] = e.description or str(e)
        exception_result["detail"] = e.detail
        exception_result["recommendation"] = e.recommendation
        exception_result["help_link"] = e.help_link
    except:
        logger.exception("Unknown error installing Jetpack")
        exception_result["status"] = JetpackStatus.Error
        exception_result["description"] = "Unexpected error installing Jetpack"
        exception_result["unhandled"] = True
        exception_result["detail"] = traceback.format_exc()
    finally:
        default_recommendation = "Check %s for more information" % logfile

        if exception_result:
            # use the caught exception
            result = exception_result

        # this is used to signal we should include default error information
        if result.get("unhandled"):
            result["recommendation"] = default_recommendation
        
        if result.get("status") == JetpackStatus.Error and "message" not in result:
            result["message"] = "Could not install Jetpack"

        # clean up any attributes used to communicate from the child process (eg unhandled),
        # as well as null values
        valid = {"status", "message", "description", "detail", "recommendation", "help_link"}
        result = {k: v for k, v in result.items() if k in valid and v != None}

    logger.info("Result: {}".format(json.dumps(result)))
    sys.stdout.write(json.dumps(result))
    sys.stdout.write("\n")
    if result.get("status", "").lower() != "ok":
        logger.warning("Status '{}'.  Exiting with non-zero code 1.", result.get("status", "").lower())
        sys.exit(1)
