#!/usr/bin/env bash

SCRIPT_DIR="$( cd "$( dirname "${BASH_SOURCE[0]}" )" && pwd )"

# this will run the tests and report the results in json format
# any health checks run by it needs to return a non-zero code if there were errors
# a part of node health checks feature: not meant to be called separately
AZUREHPC_TESTS_PATH=$1
AZUREHPC_CONF_PATH="$(dirname "${AZUREHPC_TESTS_PATH}")/conf"
LOG_PATH=$2
PATH_TO_COMPONENT_VERSIONS=$3
HEALTHCHECKS_COMPONENT_NAME=$4
JETPACK_PYTHON_PATH=$5
VERSION_CHECKER="${SCRIPT_DIR}/version_check"

function prioritize_stderr(){
if [[ $(cat "$LOG_PATH"/jetpack_health_checks.err) ]]; then
   cat "${LOG_PATH}/jetpack_health_checks.err" | tail -n1
else
   cat "${LOG_PATH}/jetpack_health_checks.out" | tail -n1
fi
}

function generate_output(){
# make sure newlines from the error message don't make it in, since that would result in a bad json
   cat <<EOF
{
"test_path" : "${1}",
"exit_code": ${2},
"message": "${3}"
}
EOF
}

if [ ! -f "${AZUREHPC_TESTS_PATH}" ]; then
   generate_output "${AZUREHPC_TESTS_PATH}" 0 "This VM image does not support health checks."
   exit 0
fi

# disable the docker0 test that flakes because of a race between cloud-init (network setup) and custom script extension (healthchecks)
for CONF in "${AZUREHPC_CONF_PATH}"/*.conf; do sed -e '/docker0/d' "${CONF}" > "${CONF}".tmp && mv "${CONF}".tmp "${CONF}"; done

ERROR_MESSAGE=$(${JETPACK_PYTHON_PATH} "${VERSION_CHECKER}" "${PATH_TO_COMPONENT_VERSIONS}" "${HEALTHCHECKS_COMPONENT_NAME}" "v0.2.6")
CC_SUPPORTED=$?

# versions prior to v0.2.6 are not supported
if [[ $CC_SUPPORTED -ne 0 ]]; then
   generate_output "${AZUREHPC_TESTS_PATH}" 0 "${ERROR_MESSAGE}"
   exit 0
else
   "${AZUREHPC_TESTS_PATH}" -v -o "$LOG_PATH/jetpack_health_checks.log" 1> "$LOG_PATH/jetpack_health_checks.out" 2> "$LOG_PATH/jetpack_health_checks.err"
   EXIT_CODE=$?
fi

# .log captures the output from nhc directly, so there might be some useful information in stdout and stderr of the test runners
echo "You can also refer to $LOG_PATH/jetpack_health_checks.out and $LOG_PATH/jetpack_health_checks.err for more information" >> "$LOG_PATH/jetpack_health_checks.log"

if [[ $EXIT_CODE -ne 0 ]]; then
   MESSAGE=$(prioritize_stderr)
   generate_output "${AZUREHPC_TESTS_PATH}" ${EXIT_CODE} "${MESSAGE}, for more details see ${LOG_PATH}/jetpack_health_checks.log"
else
   # if we reached here, either all tests passed or the VM SKU is not supported
   MESSAGE=$(prioritize_stderr)
   if echo "${MESSAGE}" | grep -q "not supported"; then
      generate_output "${AZUREHPC_TESTS_PATH}" 0 "${MESSAGE}"
   else
      generate_output "${AZUREHPC_TESTS_PATH}" 0 "HealthChecks passed"
   fi
fi

exit ${EXIT_CODE}